/* -*- C++ -*-
 *
 * Copyright (C) 2002 Jean-Daniel Fekete
 * 
 * This file is part of MillionVis.
 * 
 * MillionVis is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 * 
 * MillionVis is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with MillionVis; see the file COPYING.  If not, write to the
 * Free Software Foundation, 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA. 
 */
#include <GL/glut.h>
//#include <GL/glu.h>
#include <infovis/tree/vector_as_tree.hpp>
#include <infovis/tree/treemap/slice_and_dice.hpp>
#include <infovis/tree/treemap/squarified.hpp>
#include <infovis/tree/treemap/drawing/gl_drawer.hpp>
#include <infovis/tree/treemap/drawing/debug_drawer.hpp>
#include <infovis/tree/treemap/drawing/pick_drawer.hpp>
#include <boost/directory.h>
#include <iostream>
#include <cmath>

using namespace infovis;
//#define DOUBLE_BUFFER

typedef vector_as_tree Tree;
typedef tree_traits<Tree>::node_descriptor node_descriptor;
typedef tree_traits<Tree>::children_iterator children_iterator;
Tree t(1);
std::vector<std::string,gc_alloc<std::string> > names;
typedef std::vector<float,gc_alloc<float,true> > WeightMap;

WeightMap wm;

int win_width, win_height;
Box treemap_box;
Box label_box;
node_descriptor labeled_node;

static void build_tree(node_descriptor parent,
		       Tree& tree, const std::string& dirname,
		       std::vector<std::string,gc_alloc<std::string> >& names,
		       std::vector<float,gc_alloc<float,true> >& wm)
{
  using namespace boost::filesystem;
  for (dir_it it(dirname); it != dir_it(); ++it) {
    if ((*it)[0] == '.')
      continue;
    node_descriptor n = add_node(parent, tree);
    names.push_back(*it);
    wm.push_back(boost::filesystem::get<size>(it));
    if (boost::filesystem::get<is_directory>(it))
      build_tree(n, tree, dirname+"/"+(*it), names, wm);
  }
}

struct compare_weight {
  const WeightMap& weight;

  compare_weight(const WeightMap& wm)
    : weight(wm) {}

  bool operator()(node_descriptor n1, node_descriptor n2) {
    return weight[n1] < weight[n2];
  }
};

typedef debug_drawer<Tree, Box, gl_drawer<Tree,Box> > Drawer;

static void display()
{
#ifdef SLICE_AND_DICE
  static treemap_slice_and_dice<
    Tree,Box,WeightMap, Drawer>
    tmsd(t, wm);
  tmsd.start();
  tmsd.visit(left_to_right, treemap_box, root(t));
  tmsd.finish();
#else
  static treemap_squarified<
    Tree,Box,WeightMap, Drawer >
    tmsq(t, wm);
  tmsq.start();
  tmsq.visit(treemap_box, root(t));
  tmsq.finish();
#endif
  gl::color(1.0, 1.0, 1.0);
  draw_box(label_box);
#ifdef DOUBLE_BUFFER
  glutSwapBuffers();
#else
  glFlush();
#endif

}

static void passive_motion(int x, int y)
{
  y = win_height - y;
  typedef debug_drawer<Tree,Box, pick_drawer<Tree,Box> > Picker;
  Picker hit;
  hit.reset(x, y);
  
#ifdef SLICE_AND_DICE
  static treemap_slice_and_dice<
    Tree,Box,WeightMap,Picker&>
    tm(t, wm, hit);
  tm.start();
  tm.visit(left_to_right, treemap_box, root(t));
  tm.finish();
#else
  static treemap_squarified<
    Tree,Box,WeightMap,Picker&>
    tm(t, wm, hit);
  tm.start();
  tm.visit(treemap_box, root(t));
  tm.finish();
#endif
  if (hit.has_picked()) {
    node_descriptor h = hit.picked_node();
    if (h != labeled_node) {
      labeled_node = h;
      gl::color(1.0, 1.0, 1.0);
      draw_box(label_box);
      gl::color(0.0, 0.0, 0.0);
      gl::raster_pos(5, 5);
      const std::string& name = names[h];
      for (std::string::const_iterator i = name.begin();
	   i != name.end(); i++) {
	glutBitmapCharacter(GLUT_BITMAP_9_BY_15, *i);
      }
#ifdef DOUBLE_BUFFER
      glDrawBuffer(GL_FRONT);
      gl::raster_pos(xmin(label_box), ymin(label_box));
      glCopyPixels(int(xmin(label_box)), int(ymin(label_box)),
		   int(width(label_box)), int(height(label_box)),
		   GL_COLOR);
      glDrawBuffer(GL_BACK);
#endif
    }
  }
  glFlush();
}

static void init()
{
  treemap_box = Box(0, 20, win_width, win_height);
  label_box = Box(0, 0, win_width, 20);
  glClearColor (1, 1, 0, 1);
  glShadeModel (GL_FLAT);
  glViewport (0, 0, win_width, win_height);
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  gluOrtho2D(0, win_width, 0, win_height);
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
  labeled_node = root(t);
}

static void reshape(int w, int h)
{
  win_width = w;
  win_height = h;
  init();
}

int main(int argc, char * argv[])
{

  std::vector<float,gc_alloc<float,true> > lwm;
  lwm.push_back(0);		// weight of root is 0

  if (argc == 1) {
    names.push_back(".");
    build_tree(root(t), t, ".", names, lwm);
  }
  else {
    names.push_back(argv[1]);
    build_tree(root(t), t, argv[1], names, lwm);
  }
  
  wm.resize(lwm.size());
  for (int i = 0; i < wm.size(); i++) {
    wm[i] = log(lwm[i] + 1.0);
  }
  sum_weights(t, wm);
  sort(t, compare_weight(wm));

  win_height = 768;
  win_width = 1024;
  glutInit(&argc, argv);
#ifdef DOUBLE_BUFFER
  glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB);
#else
  glutInitDisplayMode(GLUT_RGB);
#endif
  glutInitWindowSize (win_width, win_height); 
  glutInitWindowPosition (0, 0);
  glutCreateWindow (argv[0]);

  init();
  glutDisplayFunc(display);
  glutPassiveMotionFunc(passive_motion);
  glutReshapeFunc(reshape);
  
  glutMainLoop();
  return 0;
}
